IF !DEF(HARDWARE_INC)
HARDWARE_INC = 1

; Originally from:
; https://github.com/assemblydigest/gameboy/tree/master/part-3-libraries
; With additions harvested from elsewhere

; Gameboy Color compatibility flag.
BOOT_GBC_UNSUPPORTED EQU $00
BOOT_GBC_COMPATIBLE EQU $80
BOOT_GBC_EXCLUSIVE EQU $C0

; Super Gameboy compatibility flag.
BOOT_SGB_UNSUPPORTED EQU $00
BOOT_SGB_SUPPORTED EQU $03

; Cartridge type.
BOOT_CART_ROM EQU $00
BOOT_CART_MBC1 EQU $01
BOOT_CART_MBC1_RAM EQU $02
BOOT_CART_MBC1_RAM_BATTERY EQU $03
BOOT_CART_MBC2 EQU $05
BOOT_CART_MBC2_BATTERY EQU $06
BOOT_CART_ROM_RAM EQU $08
BOOT_CART_ROM_RAM_BATTERY EQU $09
BOOT_CART_MMM01 EQU $0B
BOOT_CART_MMM01_RAM EQU $0C
BOOT_CART_MMM01_RAM_BATTERY EQU $0D
BOOT_CART_MBC3_TIMER_BATTERY EQU $0F
BOOT_CART_MBC3_TIMER_RAM_BATTERY EQU $10
BOOT_CART_MBC3 EQU $11
BOOT_CART_MBC3_RAM EQU $12
BOOT_CART_MBC3_RAM_BATTERY EQU $13
BOOT_CART_MBC4 EQU $15
BOOT_CART_MBC4_RAM EQU $16
BOOT_CART_MBC4_RAM_BATTERY EQU $17
BOOT_CART_MBC5 EQU $19
BOOT_CART_MBC5_RAM EQU $1A
BOOT_CART_MBC5_RAM_BATTERY EQU $1B
BOOT_CART_MBC5_RUMBLE EQU $1C
BOOT_CART_MBC5_RUMBLE_RAM EQU $1D
BOOT_CART_MBC5_RUMBLE_RAM_BATTERY EQU $1E
BOOT_CART_POCKET_CAMERA EQU $FC
BOOT_CART_BANDAI_TAMA5 EQU $FD
BOOT_CART_HUC3 EQU $FE
BOOT_CART_HUC1_RAM_BATTERY EQU $FF

; Rom sizes.
BOOT_ROM_32K EQU $00
BOOT_ROM_64K EQU $01
BOOT_ROM_128K EQU $02
BOOT_ROM_256K EQU $03
BOOT_ROM_512K EQU $04
BOOT_ROM_1024K EQU $05
BOOT_ROM_2048K EQU $06
BOOT_ROM_4096K EQU $07
BOOT_ROM_1152K EQU $52
BOOT_ROM_1280K EQU $53
BOOT_ROM_1536K EQU $54

; Ram sizes.
BOOT_RAM_NONE EQU $00
BOOT_RAM_2K EQU $01
BOOT_RAM_8K EQU $02
BOOT_RAM_32K EQU $03

; Destination code.
BOOT_DEST_JAPAN EQU $00
BOOT_DEST_INTERNATIONAL EQU $01

; LCD
gb_lcd_ctrl EQU $FF40
gb_lcd_stat EQU $FF41
gb_lcd_dma EQU $FF46

GB_LCD_CTRL_ON EQU $80
GB_LCD_CTRL_WIN_MAP2 EQU $40
GB_LCD_CTRL_WIN_MAP1 EQU $00
GB_LCD_CTRL_WIN_ON EQU $20
GB_LCD_CTRL_TILESET2 EQU $10
GB_LCD_CTRL_TILESET1 EQU $00
GB_LCD_CTRL_BG_MAP2 EQU $08
GB_LCD_CTRL_BG_MAP1 EQU $00
GB_LCD_CTRL_OBJ_8x16 EQU $04
GB_LCD_CTRL_OBJ_8x8 EQU $00
GB_LCD_CTRL_OBJ_ON EQU $02
GB_LCD_CTRL_BG_ON EQU $01

GB_LCD_CTRL_BIT_BG EQU 0
GB_LCD_CTRL_BIT_OBJ EQU 1
GB_LCD_CTRL_BIT_OBJ_SIZE EQU 2
GB_LCD_CTRL_BIT_BG_MAP EQU 3
GB_LCD_CTRL_BIT_TILESET EQU 4
GB_LCD_CTRL_BIT_WIN EQU 5
GB_LCD_CTRL_BIT_WIN_MAP EQU 6
GB_LCD_CTRL_BIT_ENABLE EQU 7

GB_LCD_STAT_SCANLINE_ENABLE EQU $40
GB_LCD_STAT_OAM_ENABLE EQU $20
GB_LCD_STAT_VBLANK_ENABLE EQU $10
GB_LCD_STAT_HBLANK_ENABLE EQU $08
GB_LCD_STAT_SCANLINE_FLAG EQU $04
GB_LCD_STAT_MODE_MASK EQU $03
GB_LCD_STAT_MODE_HBLANK EQU $00
GB_LCD_STAT_MODE_VBLANK EQU $01
GB_LCD_STAT_MODE_BUSY_OAM EQU $02
GB_LCD_STAT_MODE_BUSY_TRANSFER EQU $03

; Various RAM areas.
gb_ram_tile EQU $8000
gb_ram_map EQU $9800
gb_ram_map2 EQU $9C00
gb_ram_obj EQU $FE00
gb_ram_wave EQU $FF30

; Scroll - Wrapped coordinates
gb_scroll_y EQU $FF42 ; 0..255
gb_scroll_x EQU $FF43 ; 0..255

; Scanline detection.
gb_scanline_current EQU $FF44 ; current = 0 .. 153, where 144..153 indicates vblank 
gb_scanline_trigger EQU $FF45

; Window - Unwrapped coordinates.
gb_window_y EQU $FF4A ; y = 0..143
gb_window_x EQU $FF4B ; x - 7 = 0 .. 166 -- x < 7 is weird. So is fully-offscren x.

; Monochrome palette.
gb_palette_bg EQU $FF47
gb_palette_obj1 EQU $FF48
gb_palette_obj2 EQU $FF49

GB_PALETTE_COLOR_3_MASK EQU $C0
GB_PALETTE_COLOR_2_MASK EQU $30
GB_PALETTE_COLOR_1_MASK EQU $0C
GB_PALETTE_COLOR_0_MASK EQU $03
GB_PALETTE_COLOR_3_SHIFT EQU 6
GB_PALETTE_COLOR_2_SHIFT EQU 4
GB_PALETTE_COLOR_1_SHIFT EQU 2
GB_PALETTE_COLOR_0_SHIFT EQU 0

; GBC palette.
gbc_palette_bg_index EQU $FF68
gbc_palette_bg_data EQU $FF69
gbc_palette_obj_index EQU $FF6A
gbc_palette_obj_data EQU $FF6B

GBC_PALETTE_INDEX_VALUE_MASK EQU $3F
GBC_PALETTE_INDEX_AUTO_INCREMENT EQU $80

GBC_PALETTE_DATA_CHANNEL_SIZE EQU 5
GBC_PALETTE_DATA_BLUE_SHIFT EQU 10
GBC_PALETTE_DATA_GREEN_SHIFT EQU 5
GBC_PALETTE_DATA_RED_SHIFT EQU 0
GBC_PALETTE_DATA_BLUE_MASK EQU $7C00
GBC_PALETTE_DATA_GREEN_MASK EQU $03E0
GBC_PALETTE_DATA_RED_MASK EQU $001F

; GBC bank selection.
gbc_select_vram EQU $FF4F
gbc_select_wram EQU $FF70

GBC_SELECT_VRAM_BANK_0 EQU $00
GBC_SELECT_VRAM_BANK_1 EQU $01

GBC_SELECT_WRAM_BANK_MASK EQU $07

; GBC speed setting.
gbc_speed EQU $FF4D

GBC_SPEED_IS_DOUBLE EQU $80
GBC_SPEED_IS_NORMAL EQU $00
GBC_SPEED_SWITCH EQU $01

; GBC Infrared.
gbc_infrared EQU $FF5D

GBC_INFRARED_WRITE_MASK EQU $01
GBC_INFRARED_WRITE_OFF EQU $00
GBC_INFRARED_WRITE_ON EQU $01
GBC_INFRARED_READ_MASK EQU $02
GBC_INFRARED_READ_NORMAL EQU $02
GBC_INFRARED_READ_RECEIVING EQU $00
GBC_INFRARED_READ_ENABLE EQU $C0
GBC_INFRARED_READ_DISABLE EQU $00

; GBC DMA.
gbc_dma_source_high EQU $FF51
gbc_dma_source_low EQU $FF52
gbc_dma_dest_high EQU $FF53
gbc_dma_dest_low EQU $FF54
gbc_dma_transfer EQU $FF55

GBC_DMA_TRANSFER_GENERAL EQU $00 ; write
GBC_DMA_TRANSFER_HBLANK EQU $80 ; write
GBC_DMA_TRANSFER_ACTIVE EQU $00 ; read
GBC_DMA_TRANSFER_INACTIVE EQU $80 ; read

; Super Game Boy.
SGB_PAL01 EQU $00
SGB_PAL23 EQU $08
SGB_PAL03 EQU $10
SGB_PAL12 EQU $18
SGB_ATTR_BLK EQU $20
SGB_ATTR_LIN EQU $28
SGB_ATTR_DIV EQU $30
SGB_ATTR_CHR EQU $38
SGB_SOUND EQU $40
SGB_SOU_TRN EQU $48
SGB_PAL_SET EQU $50
SGB_PAL_TRN EQU $58
SGB_ATRC_EN EQU $60
SGB_TEST_EN EQU $68
SGB_ICON_EN EQU $70
SGB_DATA_SND EQU $78
SGB_DATA_TRN EQU $80
SGB_MLT_REQ EQU $88
SGB_JUMP EQU $90
SGB_CHR_TRN EQU $98
SGB_PCT_TRN EQU $A0
SGB_ATTR_TRN EQU $A8
SGB_ATTR_SET EQU $B0
SGB_MASK_EN EQU $B8
SGB_OBJ_TRN EQU $C0

; Audio volume.
gb_audio_volume EQU $FF24

GB_AUDIO_VOLUME_RIGHT_VIN EQU $80
GB_AUDIO_VOLUME_RIGHT_MASK EQU $70
GB_AUDIO_VOLUME_LEFT_VIN EQU $08
GB_AUDIO_VOLUME_LEFT_MASK EQU $07

; Audio panning.
gb_audio_pan EQU $FF25

GB_AUDIO_PAN_NOISE_R EQU $80
GB_AUDIO_PAN_NOISE_L EQU $08
GB_AUDIO_PAN_WAVE_R EQU $40
GB_AUDIO_PAN_WAVE_L EQU $04
GB_AUDIO_PAN_TONE2_R EQU $20
GB_AUDIO_PAN_TONE2_L EQU $02
GB_AUDIO_PAN_TONE_R EQU $10
GB_AUDIO_PAN_TONE_L EQU $01

; Audio toggle.
gb_audio_toggle EQU $FF26

GB_AUDIO_TOGGLE_ENABLE EQU $80
GB_AUDIO_TOGGLE_DISABLE EQU $00
GB_AUDIO_TOGGLE_NOISE EQU $08
GB_AUDIO_TOGGLE_WAVE EQU $04
GB_AUDIO_TOGGLE_TONE2 EQU $02
GB_AUDIO_TOGGLE_TONE EQU $01

; Tone 1 channel
gb_tone_sweep EQU $FF10
gb_tone_pattern EQU $FF11
gb_tone_envelope EQU $FF12
gb_tone_freq_low EQU $FF13
gb_tone_freq_high EQU $FF14

GB_TONE_SWEEP_TIME_MASK EQU $70
GB_TONE_SWEEP_DIRECTION_INC EQU $00
GB_TONE_SWEEP_DIRECTION_DEC EQU $08
GB_TONE_SWEEP_SHIFT_MASK EQU $07

GB_TONE_PATTERN_DUTY_MASK EQU $C0
GB_TONE_PATTERN_LENGTH_MASK EQU $1F

GB_TONE_ENVELOPE_INITIAL_MASK EQU $F0
GB_TONE_ENVELOPE_DIRECTION_INC EQU $08
GB_TONE_ENVELOPE_DIRECTION_DEC EQU $00
GB_TONE_ENVELOPE_SWEEP_MASK EQU $07

GB_TONE_FREQ_HIGH_RESTART EQU $80
GB_TONE_FREQ_HIGH_COUNT EQU $40
GB_TONE_FREQ_HIGH_MASK EQU $07

; Tone 2 channel
gb_tone2_pattern EQU $FF16
gb_tone2_envelope EQU $FF17
gb_tone2_freq_low EQU $FF18
gb_tone2_freq_high EQU $FF19

GB_TONE2_PATTERN_DUTY_MASK EQU $C0
GB_TONE2_PATTERN_LENGTH_MASK EQU $1F

GB_TONE2_ENVELOPE_INITIAL_MASK EQU $F0
GB_TONE2_ENVELOPE_DIRECTION_INC EQU $08
GB_TONE2_ENVELOPE_DIRECTION_DEC EQU $00
GB_TONE2_ENVELOPE_SWEEP_MASK EQU $07

GB_TONE2_FREQ_HIGH_RESTART EQU $80
GB_TONE2_FREQ_HIGH_COUNT EQU $40
GB_TONE2_FREQ_HIGH_MASK EQU $07

; Wave channel
gb_wave_toggle EQU $FF1A
gb_wave_length EQU $FF1B
gb_wave_volume EQU $FF1C
gb_wave_freq_low EQU $FF1D
gb_wave_freq_high EQU $FF1E

GB_WAVE_TOGGLE_ENABLE EQU $80
GB_WAVE_TOGGLE_DISABLE EQU $00

GB_WAVE_VOLUME_0 EQU $00
GB_WAVE_VOLUME_100 EQU $20
GB_WAVE_VOLUME_50 EQU $40
GB_WAVE_VOLUME_25 EQU $60

GB_WAVE_FREQ_HIGH_RESTART EQU $80
GB_WAVE_FREQ_HIGH_COUNT EQU $40
GB_WAVE_FREQ_HIGH_MASK EQU $07

; Noise channel
gb_noise_length EQU $FF20
gb_noise_envelope EQU $FF21
gb_noise_poly EQU $FF22
gb_noise_play EQU $FF23

GB_NOISE_LENGTH_MASH EQU $1F

GB_NOISE_ENVELOPE_INITIAL_MASK EQU $F0
GB_NOISE_ENVELOPE_DIRECTION_INC EQU $08
GB_NOISE_ENVELOPE_DIRECTION_DEC EQU $00
GB_NOISE_ENVELOPE_SWEEP_MASK EQU $07

GB_NOISE_POLY_CLOCK EQU $F0
GB_NOISE_POLY_STEP EQU $08
GB_NOISE_POLY_RATIO EQU $07

GB_NOISE_PLAY_RESTART EQU $80
GB_NOISE_PLAY_COUNT EQU $40

; Joypad.
gb_joy_ctrl EQU $FF00

GB_JOY_SELECT_FIRST_PART EQU $20
GB_JOY_SELECT_SECOND_PART EQU $10

; These bits are 0 when the appropriate button is pressed.
; Some bits have been shifted left by 4 for `readJoypad`
GB_JOY_BIT_DOWN EQU 7
GB_JOY_BIT_UP EQU 6
GB_JOY_BIT_LEFT EQU 5
GB_JOY_BIT_RIGHT EQU 4
GB_JOY_BIT_START EQU 3
GB_JOY_BIT_SELECT EQU 2
GB_JOY_BIT_B EQU 1
GB_JOY_BIT_A EQU 0

GB_JOY_DOWN EQU 1 << 7
GB_JOY_UP EQU 1 << 6
GB_JOY_LEFT EQU 1 << 5
GB_JOY_RIGHT EQU 1 << 4
GB_JOY_START EQU 1 << 3
GB_JOY_SELECT EQU 1 << 2
GB_JOY_B EQU 1 << 1
GB_JOY_A EQU 1 << 0

; Link cable.
gb_link_data EQU $FF01
gb_link_ctrl EQU $FF02

GB_LINK_CTRL_TRANSFER_START EQU $80
GB_LINK_CTRL_TRANSFER_OFF EQU $00
GB_LINK_CTRL_GBC_FAST EQU $02
GB_LINK_CTRL_GBC_NORMAL EQU $00
GB_LINK_CTRL_CLOCK_INTERNAL EQU $01
GB_LINK_CTRL_CLOCK_EXTERNAL EQU $00

; Timer.
gb_timer_divider EQU $FF04
gb_timer_counter EQU $FF05
gb_timer_modulo EQU $FF06
gb_timer_ctrl EQU $FF07

GB_TIMER_CTRL_STOP EQU $00
GB_TIMER_CTRL_START EQU $04
GB_TIMER_CTRL_CLOCK_4096 EQU $00
GB_TIMER_CTRL_CLOCK_262144 EQU $01
GB_TIMER_CTRL_CLOCK_65536 EQU $02
GB_TIMER_CTRL_CLOCK_16384 EQU $03

; Interrupts.
gb_interrupt_enable EQU $FFFF
gb_interrupt_flag EQU $FF0F

GB_INTERRUPT_VBLANK EQU $01
GB_INTERRUPT_STAT EQU $02
GB_INTERRUPT_TIMER EQU $04
GB_INTERRUPT_SERIAL EQU $04
GB_INTERRUPT_JOYPAD EQU $08

ENDC